//-----------------------------------------------------------
//    cast_and_call
//-----------------------------------------------------------
#include <tuple>
#include <type_traits>

#include "fwd.h"

namespace sykes{

    //-----------------------------------------------------------
    //    cast_and_call_helper
    //-----------------------------------------------------------
    template<std::size_t N, typename Fsig, typename Caster>
    struct cast_and_call_helper;
    
    template<std::size_t N, typename Result, typename ... Args1>
    struct cast_and_call_helper<N, Result(Args1...), static_caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            typename std::tuple_element<N - 1, typename std::remove_reference<Tuple>::type>::type
                a1 = std::get<N - 1>(tup);
            typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                a2 = static_cast<typename std::tuple_element<N - 1, std::tuple<Args1...>>::type>(a1);
            return cast_and_call_helper<N - 1, Result(Args1...), static_caster
                >::template execute<
                    F, Tuple,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type,
                    Args2...
                >(
                    func,
                    tup,
                    a2,
                    args...);
        }
    };
    
    template<std::size_t N, typename Result, typename ... Args1>
    struct cast_and_call_helper<N, Result(Args1...), dynamic_caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            typename std::tuple_element<N - 1, typename std::remove_reference<Tuple>::type>::type
                a1 = std::get<N - 1>(tup);
            typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                a2 = dynamic_cast<typename std::tuple_element<N - 1, std::tuple<Args1...>>::type>(a1);
            return cast_and_call_helper<N - 1, Result(Args1...), dynamic_caster
                >::template execute<
                    F, Tuple,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type,
                    Args2...
                >(
                    func,
                    tup,
                    a2,
                    args...);
            }
    };
    
    template<std::size_t N, typename Result, typename ... Args1, typename Caster>
    struct cast_and_call_helper<N, Result(Args1...), Caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            typename std::tuple_element<N - 1, typename std::remove_reference<Tuple>::type>::type
                a1 = std::get<N - 1>(tup);
            typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                a2 = Caster::template cast<typename std::tuple_element<N - 1, std::tuple<Args1...>>::type>(a1);
            return cast_and_call_helper<N - 1, Result(Args1...), Caster
                >::template execute<
                    F, Tuple,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type,
                    Args2...
                >
                (
                    func,
                    tup,
                    a2,
                    args...);
        }
    };
    
    template<typename Result, typename ... Args1>
    struct cast_and_call_helper<0, Result(Args1...), static_caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            return func(args...);
        }
    };
    
    template<typename Result, typename ... Args1>
    struct cast_and_call_helper<0, Result(Args1...), dynamic_caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            return func(args...);
        }
    };
    
    template<typename Result, typename ... Args1, typename Caster>
    struct cast_and_call_helper<0, Result(Args1...), Caster>
    {
        template<typename F, typename Tuple, typename ... Args2>
        static Result execute(F func, Tuple tup, Args2 ... args)
        {
            return func(args...);
        }
    };
    
    //-----------------------------------------------------------
    //    cast_and_call
    //-----------------------------------------------------------
    template<typename Fbase, typename Fcall, typename Caster>
    struct cast_and_call;
    
    template<typename Result, typename ... Args1, typename ... Args2>
    struct cast_and_call<Result(Args1...), Result(Args2...), static_caster>
    {
        template<typename Func>
        static Result execute(Func f, Args2... args)
        {
            static std::size_t const N = std::tuple_size<std::tuple<Args1...>>::value;
            static_assert(
                N == std::tuple_size<std::tuple<Args2...>>::value,
                "struct cast_and_call");
            
            std::tuple<Args2...> tup(args...);
            
            return cast_and_call_helper<N - 1, Result(Args1...), static_caster
                >::template execute<
                    Func,
                    std::tuple<Args2...>&,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                >(
                    f,
                    tup,
                    static_cast<
                        typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                    >(std::get<N - 1>(tup)));
        }
    };
    
    template<typename Result, typename ... Args1, typename ... Args2>
    struct cast_and_call<Result(Args1...), Result(Args2...), dynamic_caster>
    {
        template<typename Func>
        static Result execute(Func f, Args2... args)
        {
            static std::size_t const N = std::tuple_size<std::tuple<Args1...>>::value;
            static_assert(
                N == std::tuple_size<std::tuple<Args2...>>::value,
                "struct cast_and_call");
            
            std::tuple<Args2...> tup(args...);
            
            return cast_and_call_helper<N - 1, Result(Args1...), dynamic_caster
                >::template execute<
                    Func, std::tuple<Args2...>&,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                >(
                    f,
                    tup,
                    dynamic_cast<
                        typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                    >(std::get<N - 1>(tup)));
        }
    };
    
    template<typename Result, typename ... Args1, typename ... Args2, typename Caster>
    struct cast_and_call<Result(Args1...), Result(Args2...), Caster>
    {
        template<typename Func>
        static Result execute(Func f, Args2... args)
        {
            static std::size_t const N = std::tuple_size<std::tuple<Args1...>>::value;
            static_assert(
                N == std::tuple_size<std::tuple<Args2...>>::value,
                "struct cast_and_call");
            
            std::tuple<Args2...> tup(args...);
            
            return cast_and_call_helper<N - 1, Result(Args1...), Caster
                >::template execute<
                    Func, std::tuple<Args2...>&,
                    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                >(
                    f,
                    tup,
                    Caster::template cast<
                        typename std::tuple_element<N - 1, std::tuple<Args1...>>::type
                    >(std::get<N - 1>(tup)));
        }
    };
}//----


