
#ifndef MULTI_METHOD_DETAIL_H
#define MULTI_METHOD_DETAIL_H

#include <tuple>
#include <utility>
#include <type_traits>

#include "utility.h"
#include "vector_c.h"
#include "fwd.h"
#include "cast_and_call.h"

namespace sykes{
namespace multi_method{ namespace detail{
    
    //-----------------------------------------------------------
    //    _make_key
    //-----------------------------------------------------------
    template<std::size_t N, typename Vec, typename ... IdTps>
    struct _make_key_helper;
    
    template<typename Vec, typename IdTp1, typename ... IdTps>
    typename std::enable_if<
            (vector_c_size<Vec>::value != 1),
            typename key<typename IdTp1::id_type, vector_c_size<Vec>::value>::type
        >::type
    _make_key()
    {
        return vector_c_to_tuple<
                typename _make_key_helper<vector_c_size<Vec>::value, Vec, IdTp1, IdTps...>::type
            >::get();
    }
    
    template<typename Vec, typename IdTp1, typename ... IdTps>
    typename std::enable_if<
        (vector_c_size<Vec>::value == 1),
        typename IdTp1::id_type
    >::type
    _make_key()
    {
        return std::tuple_element<vector_c_at<0, Vec>::value, std::tuple<IdTp1, IdTps...>>::type::id;
    }
    
    template<std::size_t N, typename Vec, typename IdTp1, typename ... IdTps>
    struct _make_key_helper<N, Vec, IdTp1, IdTps...>
    {
        typedef typename
            vector_c_push_back<
                typename _make_key_helper<N - 1, Vec, IdTp1, IdTps...>::type,
                constant_value<
                    typename IdTp1::id_type,
                    std::tuple_element<
                        vector_c_at<N - 1, Vec>::value,
                        std::tuple<IdTp1, IdTps...>
                    >::type::id
                >
            >::type type;
    };
    
    template<typename Vui, typename IdTp1, typename ... IdTps>
    struct _make_key_helper<1, Vui, IdTp1, IdTps...>
    {
        typedef vector_c<
            typename IdTp1::id_type,
            std::tuple_element<
                vector_c_at<0, Vui>::value,
                std::tuple<IdTp1, IdTps...>
            >::type::id
            > type;
    };

    //-----------------------------------------------------------
    //    _make_arg_type
    //-----------------------------------------------------------
    template<std::size_t N, typename Vui, typename ... IdTps>
    struct _make_arg_type_helper;
    
    template<typename Vui, typename ... IdTps>
    struct _make_arg_type
    {
        typedef typename
            _make_arg_type_helper<
                vector_c_size<Vui>::value,
                Vui, IdTps...
            >::type type;
    };
    
    template<std::size_t N, typename Vui, typename ... IdTps>
    struct _make_arg_type_helper
    {
        typedef typename
            _tuple_push_front<
                typename _make_arg_type_helper<N - 1, Vui, IdTps...>::type,
                typename std::tuple_element<
                    vector_c_at<
                        vector_c_size<Vui>::value - N,
                        Vui
                    >::value,
                    std::tuple<IdTps...>
                >::type::type
            >::type type;
    };
    
    template<typename Vui, typename ... IdTps>
    struct _make_arg_type_helper<1, Vui, IdTps...>
    {
        typedef
            std::tuple<
                typename
                std::tuple_element<
                    vector_c_at<
                        vector_c_size<Vui>::value - 1,
                        Vui
                    >::value,
                    std::tuple<IdTps...>
                >::type::type
            > type;
    };

    //-----------------------------------------------------------
    //    _make_function
    //-----------------------------------------------------------
    template<typename Fsig, typename Fcall, typename Fbase>
    class _fBinder;
    
    template<typename Result, typename ... Args, typename Fcall, typename Fbase>
    class _fBinder<Result(Args...), Fcall, Fbase>
    {
    public:
        _fBinder(Fcall* f, Fbase* a)
            : m_func(f), m_base_func(a)
        {
        }
        
        Result operator()(Args ... args) const
        {
            return (*m_func)(m_base_func, args...);
        }
    private:
        Fcall* m_func;
        Fbase* m_base_func;
    };
    
    template<typename Fsig, typename Fcall, typename Fbase>
    _fBinder<Fsig, Fcall, Fbase>
    _fBind(Fcall* f, Fbase* a)
    {
        return _fBinder<Fsig, Fcall, Fbase>(f, a);
    }
    
    template<typename Fsig, typename Visitor>
    struct _get_function;
    
    template<typename Fcall, typename Fbase, typename Visitor, typename Caster>
    struct _make_function;
    
    template<
        typename Result,
        typename ... Args1,
        typename ... Args2,
        typename Visitor,
        typename Caster
    >
    struct _make_function<Result(Args1...), Result(Args2...), Visitor, Caster>
    {
        static std::function<Result(Args1...)>
        execute()
        {
            return
                std::function<Result(Args1...)>(
                    _fBind<Result(Args1...)>(
                        &(cast_and_call<Result(Args2...), Result(Args1...), Caster>::template execute<Result(Args2 ...)>),
                        &(Visitor::template execute<Args2...>)
                    )
                );
        }
    };
    
    //-----------------------------------------------------------
    //    _increment_mrange
    //-----------------------------------------------------------
    template<typename Its, typename Begins, typename Ends>
    struct _increment_mrange;
    
    template<std::size_t N, typename Its, typename Begins, typename Ends>
    struct _increment_mrange_helper;
    
    template<std::size_t N, std::size_t N1, std::size_t N2, typename Its, typename Begins, typename Ends>
    struct _increment_mrange_helper2;
    
    template<std::size_t ... N1, std::size_t ... N2, std::size_t ... N3>
    struct _increment_mrange<vector_c<std::size_t, N1...>, vector_c<std::size_t, N2...>, vector_c<std::size_t, N3...>>
    {
        static_assert(
            (vector_c_size<vector_c<std::size_t, N1...>>::value == vector_c_size<vector_c<std::size_t, N2...>>::value)
            && (vector_c_size<vector_c<std::size_t, N2...>>::value == vector_c_size<vector_c<std::size_t, N3...>>::value),
            "struct sykes::multi_method::detail::_increment_mrange");
        
        typedef typename
            _increment_mrange_helper<
                vector_c_size<vector_c<std::size_t, N1...>>::value - 1,
                vector_c<std::size_t, N1...>,
                vector_c<std::size_t, N2...>,
                vector_c<std::size_t, N3...>
            >::type type;
    };
    
    template<std::size_t N, std::size_t ... N1, std::size_t ... N2, std::size_t ... N3>
    struct _increment_mrange_helper<
        N, vector_c<std::size_t, N1...>,
        vector_c<std::size_t, N2...>,
        vector_c<std::size_t, N3...>
    >
    {
        static_assert(
            (N < vector_c_size<vector_c<std::size_t, N1...>>::value),
            "struct sykes::multi_method::detail::_increment_mrange_helper");
        
        typedef typename
            _increment_mrange_helper2<
                N,
                vector_c_at<N, vector_c<std::size_t, N1...>>::value,
                vector_c_at<N, vector_c<std::size_t, N3...>>::value - 1,
                vector_c<std::size_t, N1...>,
                vector_c<std::size_t, N2...>,
                vector_c<std::size_t, N3...>
            >::type type;
    };
    
    template<std::size_t ... N1, std::size_t ... N2, std::size_t ... N3>
    struct _increment_mrange_helper<
        0,
        vector_c<std::size_t, N1...>,
        vector_c<std::size_t, N2...>,
        vector_c<std::size_t, N3...>
    >
    {
        typedef typename
            vector_c_increment_at<0, vector_c<std::size_t, N1...>>::type type;
    };
    
    template<
        std::size_t N,
        std::size_t N1,
        std::size_t N2,
        std::size_t ... Nit,
        std::size_t ... Nbeg,
        std::size_t ... Nend
    >
    struct _increment_mrange_helper2<
        N, N1, N2,
        vector_c<std::size_t, Nit...>,
        vector_c<std::size_t, Nbeg...>,
        vector_c<std::size_t, Nend...>
    >
    {
        static_assert(
            (N < vector_c_size<vector_c<std::size_t, Nit...>>::value),
            "struct sykes::multi_method::detail::_increment_mrange_helper");
        typedef typename
            vector_c_increment_at<N, vector_c<std::size_t, Nit...>>::type type;
    };
    
    template<
        std::size_t N,
        std::size_t N1,
        std::size_t ... Nit,
        std::size_t ... Nbeg,
        std::size_t ... Nend
    >
    struct _increment_mrange_helper2<
        N, N1, N1,
        vector_c<std::size_t, Nit...>,
        vector_c<std::size_t, Nbeg...>,
        vector_c<std::size_t, Nend...>
    >
    {
        static_assert(
            (N < vector_c_size<vector_c<std::size_t, Nit...>>::value),
            "struct sykes::multi_method::detail::_increment_mrange_helper");
        typedef typename
            _increment_mrange_helper2<
                N - 1,
                vector_c_at<N - 1, vector_c<std::size_t, Nit...>>::value,
                vector_c_at<N - 1, vector_c<std::size_t, Nend...>>::value - 1,
                typename vector_c_replace_at<
                    N,
                    vector_c<std::size_t, Nit...>,
                    constant_value<
                        std::size_t,
                        vector_c_at<N, vector_c<std::size_t, Nbeg...>>::value
                    >
                >::type,
                vector_c<std::size_t, Nbeg...>,
                vector_c<std::size_t, Nend...>
            >::type type;
    };
    
    template<
        std::size_t N1,
        std::size_t ... Nit,
        std::size_t ... Nbeg,
        std::size_t ... Nend
    >
    struct _increment_mrange_helper2<
        0, N1, N1,
        vector_c<std::size_t, Nit...>,
        vector_c<std::size_t, Nbeg...>,
        vector_c<std::size_t, Nend...>
    >
    {
        typedef typename
            vector_c_increment_at<0, vector_c<std::size_t, Nit...>>::type type;
    };
    
    
    //-----------------------------------------------------------
    //    _insert_element
    //-----------------------------------------------------------
    template<
        typename Fsig,
        typename Func,
        typename Map,
        typename Its,
        typename Begins,
        typename Ends,
        typename ... Id_and_types
    >
    struct _insert_elements;
    
    template<
        typename Result,
        typename ... BaseArgs,
        typename Func,
        typename Map,
        typename Caster,
        std::size_t ... Nits,
        std::size_t ... Nbeg,
        std::size_t ... Nend,
        typename IdTp1,
        typename ... Id_and_Types
    >
    struct _insert_elements<
        Result(BaseArgs...),
        Func,
        Map,
        Caster,
        vector_c<std::size_t, Nits...>,
        vector_c<std::size_t, Nbeg...>,
        vector_c<std::size_t, Nend...>,
        IdTp1, Id_and_Types...
    >
    {
        static void execute(Map& _map)
        {
            _map.insert(
                std::pair<
                    typename key<typename IdTp1::id_type, std::tuple_size<std::tuple<BaseArgs...>>::value>::type const,
                    std::function<Result(BaseArgs...)>
                >(
                    _make_key<vector_c<std::size_t, Nits...>, IdTp1, Id_and_Types...>(),
                    _make_function<
                        Result(BaseArgs...),
                        typename _make_function_type<
                            Result,
                            typename _make_arg_type<
                                vector_c<std::size_t, Nits...>, IdTp1, Id_and_Types...
                            >::type
                        >::type,
                        Func,
                        Caster
                    >::execute()
                )
            );
            
            _insert_elements<
                Result(BaseArgs...),
                Func,
                Map,
                Caster,
                typename
                _increment_mrange<
                    vector_c<std::size_t, Nits...>, vector_c<std::size_t, Nbeg...>, vector_c<std::size_t, Nend...>
                >::type,
                vector_c<std::size_t, Nbeg...>,
                vector_c<std::size_t, Nend...>,
                IdTp1, Id_and_Types...
            >::execute(_map);
            
        }
    };
    
    template<
        typename Result,
        typename ... BaseArgs,
        typename Func,
        typename Map,
        typename Caster,
        std::size_t N,
        std::size_t ... N1,
        std::size_t ... N2,
        std::size_t ... N3,
        typename IdTp1,
        typename ... Id_and_Types
    >
    struct _insert_elements<
        Result(BaseArgs...),
        Func,
        Map,
        Caster,
        vector_c<std::size_t, N, N1...>,
        vector_c<std::size_t, N2...>,
        vector_c<std::size_t, N, N3...>,
        IdTp1,
        Id_and_Types ...
    >
    {
        static void execute(Map& _map)
        {
        }
    };
}}}//---- namespace

#endif

