
#ifndef TUPLE_UTILITY_H
#define TUPLE_UTILITY_H

#include <tuple>

namespace sykes{
    // _tuple_push_front
    template<typename Tuple, typename Tp>
    struct _tuple_push_front;

    template<typename T1, typename T2>
    struct _tuple_push_front<std::tuple<T1>, T2>
    {
        typedef std::tuple<T2, T1> type;
    };

    template<typename T1, typename T2, typename T3, typename ... Types>
    struct _tuple_push_front<std::tuple<T1, T2, Types...>, T3>
    {
        typedef std::tuple<T3, T1, T2, Types...> type;
    };
    
    // _tuple_pop_front
    template<typename Tuple>
    struct _tuple_pop_front;

    template<typename T1, typename T2>
    struct _tuple_pop_front<std::tuple<T1, T2> >
    {
        typedef std::tuple<T2> type;
    };

    // tupleN
    template<std::size_t N, typename Tp>
    struct tupleN;

    template<typename Tp>
    struct tupleN<1, Tp>
    {
        typedef std::tuple<Tp> type;
    };

    template<std::size_t N, typename Tp>
    struct tupleN
    {
        typedef typename
            _tuple_push_front<
                typename tupleN<N - 1, Tp>::type, Tp
            >::type type;
    };
    
    
    template<typename Fsig>
    struct _result_type;
    
    template<typename R, typename ... Args>
    struct _result_type<R(Args...)>
    {
        typedef R type;
    };
    
    template<typename Fsig>
    struct _arg_type;
    
    template<typename R, typename ... Args>
    struct _arg_type<R(Args...)>
    {
        typedef std::tuple<Args...> type;
    };
    
    template<typename R, typename ArgTuple>
    struct _make_function_type;
    
    template<typename R, typename ... Args>
    struct _make_function_type<R, std::tuple<Args...>>
    {
        typedef R type(Args...);
    };
    
    // tuple_to_arg
    template<std::size_t N, typename Fsig>
    struct tuple_to_arg_helper;
    
    template<std::size_t N, typename Result, typename ... Args1>
    struct tuple_to_arg_helper<N, Result(Args1...)>
    {
	    template<typename Func, typename Tuple, typename ... Args2>
	    static Result process(Func _func, Tuple& _tup, Args2 ... _args)
	    {
		    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type a = std::get<N - 1>(_tup);
		    return tuple_to_arg_helper<N - 1, Result(Args1...)>::process(_func, _tup, a, _args...);
	    }
	    
	    template<typename Func, typename Tuple, typename ... Args2>
	    static Result process(Func _func, Tuple const& _tup, Args2 ... _args)
	    {
		    typename std::tuple_element<N - 1, std::tuple<Args1...>>::type a = std::get<N - 1>(_tup);
		    return tuple_to_arg_helper<N - 1, Result(Args1...)>::process(_func, _tup, a, _args...);
	    }
    };

    template<typename Result, typename ... Args1>
    struct tuple_to_arg_helper<0, Result(Args1...)>
    {
	    template<typename Func, typename Tuple, typename ... Args2>
	    static Result process(Func _func, Tuple& _tup, Args2 ... _args)
	    { return _func(_args...); }
	    
	    template<typename Func, typename Tuple, typename ... Args2>
	    static Result process(Func _func, Tuple const& _tup, Args2 ... _args)
	    { return _func(_args...); }
	    
	    template<typename Func, typename Tuple>
	    static Result process(Func _func, Tuple& _tup)
	    { return _func(); }
	    
	    template<typename Func, typename Tuple>
	    static Result process(Func _func, Tuple const& _tup)
	    { return _func(); }
    };

    template<typename Fsig, typename Func, typename Tuple>
    typename _result_type<Fsig>::type
    tuple_to_arg(Func _func, Tuple& _tup)
    {
	    return tuple_to_arg_helper<std::tuple_size<Tuple>::value, Fsig>::process(_func, _tup);
    }
    
    template<typename Fsig, typename Func, typename Tuple>
    typename _result_type<Fsig>::type
    tuple_to_arg(Func _func, Tuple const& _tup)
    {
	    return tuple_to_arg_helper<std::tuple_size<Tuple>::value, Fsig>::process(_func, _tup);
    }
    
}//---- namespace

#endif

