//-----------------------------------------------------------
//    vector
//-----------------------------------------------------------
#ifndef vector_c_H
#define vector_c_H

#include <cstddef>
#include <tuple>
#include <type_traits>
#include "utility.h"

namespace sykes{
    
    template<typename Tp, Tp V>
    struct constant_value
    {
        typedef Tp value_type;
        static value_type const value = V;
    };
    
    //-----------------------------------------------------------
    //    vector_c
    //-----------------------------------------------------------
    template<typename Tp, Tp ... N>
    struct vector_c
    {
        typedef Tp value_type;
    };

    //-----------------------------------------------------------
    //    vector_c_size
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_size;
    
    template<typename Tp, Tp N1, Tp ... N>
    struct vector_c_size<vector_c<Tp, N1, N...>>
    {
        static std::size_t const
            value = 1 + vector_c_size<vector_c<Tp, N...>>::value;
    };
    
    template<typename Tp>
    struct vector_c_size<vector_c<Tp>>
    {
        static std::size_t const
            value = 0;
    };
    
    //-----------------------------------------------------------
    //    vector_c_at
    //-----------------------------------------------------------
    template<std::size_t N, typename Vec>
    struct vector_c_at;
    
    template<typename Tp, Tp N1, Tp ... Ns>
    struct vector_c_at<0, vector_c<Tp, N1, Ns...>>
    {
        static Tp const value = N1;
    };
    
    template<typename Tp, std::size_t N, Tp N1, Tp ... Ns>
    struct vector_c_at<N, vector_c<Tp, N1, Ns...>>
    {
        static_assert(
            (N < (vector_c_size<vector_c<Tp, N1, Ns...> >::value)),
            "struct vector_c_at");
        
        static Tp const
            value = vector_c_at<N - 1, vector_c<Tp, Ns...>>::value;
    };
    
    //-----------------------------------------------------------
    //    vector_c_pop_front push_front
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_pop_front;
    
    template<typename Vec, typename IntC>
    struct vector_c_push_front;
    
    template<typename Tp, Tp N1, Tp ... N>
    struct vector_c_pop_front<vector_c<Tp, N1, N...>>
    {
        typedef vector_c<Tp, N...> type;
    };
    
    template<typename Tp, Tp N1, Tp ... N>
    struct vector_c_push_front<vector_c<Tp, N...>, constant_value<Tp, N1>>
    {
        typedef vector_c<Tp, N1, N...> type;
    };
    
    //-----------------------------------------------------------
    //    vector_cut_in_range
    //-----------------------------------------------------------
    template<std::size_t N, typename Vec>
    struct _vector_c_cut_in_range_helper1;
    
    template<std::size_t N, typename Vec>
    struct _vector_c_cut_in_range_helper2;
    
    template<std::size_t Begin, std::size_t End, typename Vec>
    struct vector_c_cut_in_range
    {
        static_assert(
            Begin <= End
            && End <= vector_c_size<Vec>::value,
            "sykes::vector_c_cut_in_range");
        
        typedef typename
            _vector_c_cut_in_range_helper2<
                End - Begin,
                typename _vector_c_cut_in_range_helper1<Begin, Vec>::type
            >::type type;
    };
    
    template<std::size_t N, typename Vec>
    struct _vector_c_cut_in_range_helper1
    {
        typedef typename
            _vector_c_cut_in_range_helper1<
                N - 1,
                typename vector_c_pop_front<Vec>::type
            >::type type;
    };
    
    template<typename Vec>
    struct _vector_c_cut_in_range_helper1<0, Vec>
    {
        typedef Vec type;
    };
    
    template<typename Tp>
    struct _vector_c_cut_in_range_helper2<0, vector_c<Tp>>
    {
        typedef vector_c<Tp> type;
    };
    
    template<typename Tp, Tp N1, Tp ... Ns>
    struct _vector_c_cut_in_range_helper2<0, vector_c<Tp, N1, Ns...>>
    {
        typedef vector_c<Tp> type;
    };
    
    template<std::size_t N, typename Tp, Tp N1, Tp ... Ns>
    struct _vector_c_cut_in_range_helper2<N, vector_c<Tp, N1, Ns...>>
    {
        typedef typename
            vector_c_push_front<
                typename _vector_c_cut_in_range_helper2<
                    N - 1,
                    vector_c<Tp, Ns...>
                >::type,
                constant_value<Tp, N1>
            >::type type;
    };
    
    
    //-----------------------------------------------------------
    //    vector_c_joint
    //-----------------------------------------------------------
    template<typename Vec1, typename Vec2>
    struct vector_c_joint;
    
    template<std::size_t N, typename Vec1, typename Vec2>
    struct _vector_c_joint_helper
    {
        typedef typename
            _vector_c_joint_helper<
                N - 1, Vec1,
                typename vector_c_push_front<
                    Vec2,
                    constant_value<
                        typename Vec1::value_type,    
                        vector_c_at<N - 1, Vec1>::value
                    >
                >::type
            >::type type;
    };
    
    template<typename Vec1, typename Vec2>
    struct _vector_c_joint_helper<0, Vec1, Vec2>
    {
        typedef Vec2 type;
    };
    
    template<typename Tp, Tp ... N1, Tp ... N2>
    struct vector_c_joint<vector_c<Tp, N1...>, vector_c<Tp, N2...>>
    {
        typedef typename
            _vector_c_joint_helper<
                vector_c_size<vector_c<Tp, N1...>>::value,
                vector_c<Tp, N1...>, vector_c<Tp, N2...>
            >::type type;
    };
    
    //-----------------------------------------------------------
    //    vector_c_pop_back push_back
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_pop_back
    {
        typedef typename
            vector_c_cut_in_range<
                0, vector_c_size<Vec>::value - 1, Vec
            >::type type;
    };
    
    template<typename Vec, typename IntC>
    struct vector_c_push_back;
    
    template<typename Tp, Tp N1, Tp ... N>
    struct vector_c_push_back<vector_c<Tp, N...>, constant_value<Tp, N1>>
    {
        typedef typename
            vector_c_joint<
                vector_c<Tp, N...>,
                vector_c<Tp, N1>
            >::type type;
    };
    
    //-----------------------------------------------------------
    //    vector_c_replace_at
    //-----------------------------------------------------------
    template<std::size_t N1, typename Vec, typename IntC>
    struct vector_c_replace_at;
    
    template<std::size_t N1, typename Tp, Tp Nr, Tp ... Ns>
    struct vector_c_replace_at<N1, vector_c<Tp, Ns...>, constant_value<Tp, Nr>>
    {
        static_assert(
            N1 < vector_c_size<vector_c<Tp, Ns...>>::value,
            "struct vector_c_replace_at");
        
        typedef typename
            vector_c_joint<
                typename vector_c_push_back<
                    typename vector_c_cut_in_range<0, N1, vector_c<Tp, Ns...>>::type,
                    constant_value<Tp, Nr>
                >::type,
                typename vector_c_cut_in_range<
                    N1 + 1, vector_c_size<vector_c<Tp, Ns...>>::value, vector_c<Tp, Ns...>
                >::type
            >::type type;
    };
    
    //-----------------------------------------------------------
    //    vector_c_erase_at
    //-----------------------------------------------------------
    template<std::size_t N1, typename Vec>
    struct vector_c_erase_at
    {
        static_assert(
            N1 < vector_c_size<Vec>::value,
            "struct vector_c_erase_at");
        
        typedef typename
            vector_c_joint<
                typename vector_c_cut_in_range<0, N1, Vec>::type,
                typename vector_c_cut_in_range<
                    N1 + 1, vector_c_size<Vec>::value, Vec
                >::type
            >::type type;
    };
    
    //-----------------------------------------------------------
    //    vector_c_increment_at
    //-----------------------------------------------------------
    template<std::size_t N, typename Vec>
    struct vector_c_increment_at
    {
        static_assert(
            (N < (vector_c_size<Vec>::value)),
            "struct vector_c_at");
        
        typedef typename
            vector_c_replace_at<
                N, Vec,
                constant_value<
                    typename Vec::value_type,
                    (vector_c_at<N, Vec>::value + 1)
                >
            >::type type;
    };
    
    
    //-----------------------------------------------------------
    //    vector_c_to_tuple
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_to_tuple;
    
    template<typename Tp, Tp ... N>
    struct vector_c_to_tuple<vector_c<Tp, N...>>
    {
        typedef typename tupleN<
                    vector_c_size<vector_c<Tp, N...>>::value,
                    Tp
                >::type tuple_type;
        
        static tuple_type get()
        {
            typedef typename tupleN<
                    vector_c_size<vector_c<Tp, N...>>::value,
                    Tp
                >::type tuple_type;
            return tuple_type(N...);
        }
    };
    
    //-----------------------------------------------------------
    //    is_vector_c
    //-----------------------------------------------------------
    template<typename Vec>
    struct is_vector_c
    {
        static bool const value = false;
    };
    
    template<typename Tp, Tp ... N>
    struct is_vector_c<vector_c<Tp, N...>>
    {
        static bool const value = true;
    };
    
    //-----------------------------------------------------------
    //    vector_c_is_sorted
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_is_sorted;
    
    template<typename Tp, Tp N1, Tp N2, Tp ... N>
    struct vector_c_is_sorted<vector_c<Tp, N1, N2, N...>>
    {
        static bool const value = (N1 <= N2)
            && vector_c_is_sorted<vector_c<Tp, N2, N...>>::value;
    };
    
    template<typename Tp, Tp N1>
    struct vector_c_is_sorted<vector_c<Tp, N1>>
    {
        static bool const value = true;
    };
    
    template<typename Tp>
    struct vector_c_is_sorted<vector_c<Tp>>
    {
        static bool const value = true;
    };
    
    //-----------------------------------------------------------
    //    vector_c_max
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_max;
    
    template<typename Tp, Tp N1, Tp N2, Tp ... N>
    struct vector_c_max<vector_c<Tp, N1, N2, N...>>
    {
        static Tp const value
            = (N1 >= vector_c_max<vector_c<Tp, N2, N...>>::value) ?
                N1 : vector_c_max<vector_c<Tp, N2, N...>>::value;
        
        static std::size_t const position
            = (N1 >= vector_c_max<vector_c<Tp, N2, N...>>::value) ?
                0 : 1 + vector_c_max<vector_c<Tp, N2, N...>>::position;
    };
    
    template<typename Tp, Tp N>
    struct vector_c_max<vector_c<Tp, N>>
    {
        static Tp const value = N;
        static std::size_t const position = 0;
    };
    
    //-----------------------------------------------------------
    //    vector_c_min
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_min;
    
    template<typename Tp, Tp N1, Tp N2, Tp ... N>
    struct vector_c_min<vector_c<Tp, N1, N2, N...>>
    {
        static Tp const value
            = (N1 <= vector_c_min<vector_c<Tp, N2, N...>>::value) ?
                N1 : vector_c_min<vector_c<Tp, N2, N...>>::value;
        
        static std::size_t const position
            = (N1 <= vector_c_min<vector_c<Tp, N2, N...>>::value) ?
                0 : 1 + vector_c_min<vector_c<Tp, N2, N...>>::position;
    };
    
    template<typename Tp, Tp N>
    struct vector_c_min<vector_c<Tp, N>>
    {
        static Tp const value = N;
        static std::size_t const position = 0;
    };
    
    //-----------------------------------------------------------
    //    vector_c_sort
    //-----------------------------------------------------------
    template<typename Vec>
    struct vector_c_sort;
    
    template<typename Tp, Tp N1, Tp N2, Tp ... N>
    struct vector_c_sort<vector_c<Tp, N1, N2, N...>>
    {
        typedef typename
            vector_c_push_front<
                typename vector_c_sort<
                    typename
                    vector_c_erase_at<
                        vector_c_min<vector_c<Tp, N1, N2, N...>>::position,
                        vector_c<Tp, N1, N2, N...>
                    >::type
                >::type,
                constant_value<
                    Tp,
                    vector_c_min<vector_c<Tp, N1, N2, N...>>::value
                >
            >::type type;
    };
    
    template<typename Tp, Tp N>
    struct vector_c_sort<vector_c<Tp, N>>
    {
        typedef vector_c<Tp, N> type;
    };
    
    //-----------------------------------------------------------
    //    vector_c_N
    //-----------------------------------------------------------
    template<std::size_t L, typename Tp, Tp N>
    struct vector_c_N;
    
    template<std::size_t L, typename Tp, Tp N>
    struct vector_c_N
    {
        typedef typename
            vector_c_push_front<
                typename vector_c_N<L - 1, Tp, N>::type,
                constant_value<Tp, N>
            >::type type;
    };
    
    template<typename Tp, Tp N>
    struct vector_c_N<0, Tp, N>
    {
        typedef vector_c<Tp> type;
    };
}


#endif








